#Donaldson, ChE 210D final project

#import python modules
import numpy as np
import time

#import compiled fortran library
import final

import sys
import os
import atomwrite

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
rc = 2.5
N = 500
L = 10
dt = 0.001
T = 1.0
#set the random number seed; useful for debugging
#np.random.seed(342324+2)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [final.calcenergy(Pos + NormDir * x, NTypes, M, L, Cut) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = final.calcenergy(Pos + NormDir * dx, NTypes, M, L, Cut)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [final.calcenergy(Pos + NormDir * 2. * dx, NTypes, M, L, Cut)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [final.calcenergy(Pos + NormDir * Dists[-1], NTypes, M, L, Cut)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = final.calcenergy(Pos + NormDir * d3, NTypes, M, L, Cut)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = final.calcenergy(Pos + NormDir * d3, NTypes, M, L, Cut)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = final.calcenergyforces(Pos, NTypes, M, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = final.calcenergyforces(Pos, NTypes, M, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    #print "hi", NLat
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) + 0.2* L
    #print "hello", r
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = final.calcenergyforces(Pos, NTypes, M, L, Cut, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))
    
def PlaceWall(L, WallTypeOne, WallTypeTwo):
    """Returns the positions and types of the wall atoms
    Input: L, box length
    WallType1: hydrophobic (0) or hydrophilic (1)
    WallType2: hydrophobic (0) or hydrophilic (1)
    Output: WallPos, (N,4) array of the positions and types of the two walls (x, y, z, Type)
    WallType One, WallTypeTwo, WallNames
    """
    NLats = L
    LatSpacing = L / NLats
    #make an array of lattice sites
    r = LatSpacing * np.arange(NLats, dtype=float) + 0.01*L

    #make the position array of the wall atoms
    WallPos = np.zeros((2*len(r)**2,3),float)
    WallNames = []
    #Empty vector to be filled with names of the hydrophobic/hydrophilic walls
    j=0
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    for x in r:
        for y in r:
            #The Plates positions are determined here such that the plates span the x and y dimensions of the entire box
            WallPos[j]=np.array([x,y,0],float) #bottom plate at z = 0
            if WallTypeOne == 0:
                WallNames.append("O")
            else:
                WallNames.append("N")
            j+=1
            WallPos[j]=np.array([x,y,L],float) #top plate at z = L
            if WallTypeTwo == 0:
                WallNames.append("O")
            else:
                WallNames.append("N")
            j+=1
    WallPos = np.array(WallPos, float)
    return WallPos, WallTypeOne, WallTypeTwo, WallNames

def PolyTypes(N, M, Type):
    """
    Returns the types and names of the polymer molecules, allowing for
    easy control of the polymer types
    Inputs: N, total number of atoms ; M, number of atoms per polymer molecule;
    Type, the type of polymer to be used
    Outputs: NTypes and Names, length N arrays of the types of each atom (0 for hydrophobic
    and 1 for hydrophilic)
    """
    i = 0
    NTypes = []
    Holder = []
    Names = []
    x = 0
    while i < N:
        i+=1
        if Type == "block":
            if (i-1) % M < (M/2):
                NTypes.append(0)
            else:
                NTypes.append(1)
        if Type == "phobic":
            NTypes.append(0)
        if Type == "philic":
            NTypes.append(1)
        if NTypes[i-1] == 0:
            Names.append("O")
        else:
            Names.append("N")
    NTypes = np.array(NTypes, int)
    return NTypes, Names


def Histogram(Pos,L,m, NTypes):
    """Returns a histogram of the z-positions for the atoms
Input:
    Pos: (N,3) array of atomic positions
    L: Box length
    m: number of bins
    NTypes: length N list of atom types (hydrophobic or hydrophilic)
Output:
    PhobCount: Histogram of z positions for the hydroPHOBic atoms
    PhilCount: Histogram of z positions for the hydroPHILic atoms
    Bin: Bin coordinates (in the z direction)
"""
    dz=float(L)/float(m) #bin width
    Bin=[(i*dz) for i in range(m+1)]
    PhobCount = np.zeros((len(Bin)-1), int)
    PhilCount = np.zeros((len(Bin)-1), int)
    NewPos = np.zeros((len(Pos),2),dtype=float)
    NewPosZ = []
    for z in Pos:
        Posz = z[2]
        NewPosZ.append(Posz)
    NewPosZ = np.array(NewPosZ)
    NewPosZ = NewPosZ[:,np.newaxis]
    NTypes = NTypes[:,np.newaxis]
    NewPos = np.concatenate((NewPosZ,NTypes),axis=1)
    for x in NewPos: #Scroll through all the positions of Pos
        if x[1]==0.: #for hydrophobic atoms only
            k=0
            while k<(len(Bin)-1): #iterate through all the bin sets
                if x[0]>=Bin[k] and x[0]<Bin[k+1]: #check to see if the z-position is within a particular bin
                    PhobCount[k]+=1 # then add a count in bin k
                k+=1 #iterare the bin index
        else:    # hydrophilic atoms only
            k=0
            while k<(len(Bin)-1):
                if x[0]>=Bin[k] and x[0]<Bin[k+1]:
                    PhilCount[k]+=1
                k+=1
    return Bin,PhobCount,PhilCount

#### YOUR CODE HERE ####

#### YOU WILL NEED TO DEFINE THE FOLLOWING GLOBAL ####
#### VARIABLES FOR THE CODE ABOVE TO WORK         ####

RescaleFreq = 100
Mlist = [10]
for M in Mlist:
    M=10
    PdbFreq = 200 #write to Pdb file for movie viewing every PdbFreq steps
    dt = 0.001
    t = 0.
    i = 0
    NStep = 0
    NStepsEquil1 = 50000
    NStepsEquil2 = 50000
    NStepsProd = 100000
    HistFreq = 200
    NumHist = 0
    dz = L / 20.
    
    
    WallPos, WallTypeOne, WallTypeTwo, WallNames = PlaceWall(L, 1, 0) #place the walls
    NTypes, Names = PolyTypes(N, M, "phobic") # initialize the type of polymer
    AllNames = Names + WallNames
    Pdb = atomwrite.pdbfile("Phobic_PhobicWalls_Hist_testXX1.pdb", L, AtomNames = AllNames)

    T = 1.0
    AllAtomPos = []
    Pos = InitPositions(N, L)
    AllAtomPos = np.concatenate((Pos, WallPos))
    Pdb.write(AllAtomPos)
    Vel = InitVelocities(N, T)
    Accel = InitAccel(Pos)
    PE, Pos = ConjugateGradient(Pos, 0.001, 1.e-8, 1.e-10) # energy minimize the initial configuration with conj grad
    AllAtomPos = np.concatenate((Pos, WallPos))
    Pdb.write(AllAtomPos)
    TotEnergyArray = []
    NewTotEnergyArray = []
    (Bin, PhobCount, PhilCount) = Histogram(Pos, L, 20, NTypes)
    Phob_Hist=np.zeros_like(PhobCount) #Empty vector to fill with histogram values 
    Phil_Hist=np.zeros_like(PhilCount) #Empty vector to fill with histogram values
    print Bin, PhobCount,PhilCount

    while NStep <= NStepsEquil1:
        if (i % RescaleFreq == 0):
            Vel = RescaleVelocities(Vel, T)
            i = 0
        
        Pos, Vel, Accel, KEnergy, PEnergy = final.vvintegratewalls(Pos, NTypes, WallTypeOne, WallTypeTwo, Vel, Accel, M, L, Cut, dt)
        t += dt
        NStep += 1
        i += 1
        if NStep % PdbFreq == 0:
            AllAtomPos = np.concatenate((Pos, WallPos))
            Pdb.write(AllAtomPos)
    print "1st equilib"
    NStep = 0
    while NStep <= NStepsEquil2:
        if (i % RescaleFreq == 0):
            Vel = RescaleVelocities(Vel, T)
            i = 0
        NStep += 1
        i += 1
        Pos, Vel, Accel, KEnergy, PEnergy = final.vvintegratewalls(Pos, NTypes, WallTypeOne, WallTypeTwo, Vel, Accel, M, L, Cut, dt)
        TotalEnergy = KEnergy + PEnergy
        t += dt
        TotEnergyArray.append(TotalEnergy)
        if NStep % PdbFreq == 0:
            AllAtomPos = np.concatenate((Pos, WallPos))
            Pdb.write(AllAtomPos)
        if (NStep % 5000 == 0):
            print t, "            ", TotalEnergy, "              ", PEnergy
    TotEnergyArray = np.array(TotEnergyArray)
    MeanTotEnergy = np.mean(TotEnergyArray)
    TargetKE = MeanTotEnergy - PEnergy
    Vel = np.sqrt(TargetKE/KEnergy) * Vel #rescale velocities to target KE before turning off rescaling
    Pos0 = Pos.copy()
    NStep = 0
    print "Rescaling off"
    while NStep < NStepsProd:
        NStep += 1
        t += dt
        Pos, Vel, Accel, KEnergy, PEnergy = final.vvintegratewalls(Pos, NTypes, WallTypeOne, WallTypeTwo, Vel, Accel, M, L, Cut, dt)
        MSqDisp = np.mean((Pos - Pos0)**2)
        if NStep % PdbFreq == 0:
            AllAtomPos = np.concatenate((Pos, WallPos))
            Pdb.write(AllAtomPos)
        TotalEnergy = KEnergy + PEnergy
        NewTotEnergyArray.append(TotalEnergy)
        if (NStep % 5000 == 0):
            print t, "            ", TotalEnergy, "              ", PEnergy
        if NStep % HistFreq == 0: # collect histograms
            NumHist +=1
            (Bin, PhobCount, PhilCount) = Histogram(Pos, L, 20, NTypes)
            Phob_Hist = Phob_Hist + PhobCount
            Phil_Hist = Phil_Hist + PhilCount
    Phob_Hist = Phob_Hist/float(NumHist * L * L * dz)
    Phil_Hist = Phil_Hist/float(NumHist * L * L * dz)
    print Bin, Phob_Hist, Phil_Hist
    Pdb.close()